/*
* ==========================================================================
*       Filename:  antzer_faro_can_sensor.c
*    Description:  Get raw can and sensor data. If define J1939_FUNCTION, read can data by the AZ_VC_J1939_Read function.
*        Version:  1.1
*        Created:  2020.08.07
*       Compiler:  gcc
*         Author:  Vincent
*        Company:  Antzer
* ==========================================================================
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include "faro_can_sdk.h"

//#define J1939_FUNCTION  //Use J1939 function.


struct sensor_ctrl_format_t msg;
pthread_mutex_t bus_mutex;
pthread_t do_read_can_sensor_thread;
pthread_t do_request_sensor_thread;
int stop_thread = 0;

//-------------------------
void request_acc_data() {
	pthread_mutex_lock(&bus_mutex);
	int ret=0;
	// Send command for geting acceleration data
	ret = AZ_VC_GetACCData();
	if(ret != 0)
	{
		printf("ERROR: Failed to get Acceleration data");
		return;
	}
	pthread_mutex_unlock(&bus_mutex);
}

float lsm303agr_from_fs_2g_hr_to_mg(short lsb)
{
	return ((float)lsb / 16.0f) * 0.98f;
}

//-------------------------
void get_sensor_data() {

	pthread_mutex_lock(&bus_mutex);
	int i;
	time_t mytime = time(NULL);
	char * time_str = ctime(&mytime);
	time_str[strlen(time_str)-1] = '\0';

	int ret=0;
	int read_size = 0;
	unsigned short utemp;
	short int x, y, z;
	float acc_x, acc_y, acc_z;
	float fTemp;
	struct sensor_ctrl_format_t msg;

	memset(&msg, 0x00, sizeof(msg));
	msg.cmd = 0x54;
	msg.dlc_cfg = 0x05;
	msg.ctrl1 = 0x00;
	msg.ctrl2 = 0x00;
	msg.ctrl3 = 0x00;
	msg.ctrl4 = 0X00;
	msg.ctrl5 = 0x00;
	msg.ctrl6 = 0x00;

	// Read sensor data
	for(i = 0; i < AZ_VC_Sensor_Read_Amount(); i++) //The amount of sensor data received.
	{
		ret = AZ_VC_Sensor_Read(&msg, &read_size);
	
		if(ret != 0)
		{
			printf("ERROR: Failed to read sensor data.");
			return;
		}
		
		utemp = (msg.ctrl2 << 8) + (msg.ctrl1);
		x = (short int)utemp;
		utemp = (msg.ctrl4 << 8) + (msg.ctrl3);
		y = (short int)utemp;
		utemp = (msg.ctrl6 << 8) + (msg.ctrl5);
		z = (short int)utemp;

		printf("%s::,Accelerometer raw data: X :%d, Y:%d, Z:%d\n",time_str,x,y,z);
		
		fTemp = lsm303agr_from_fs_2g_hr_to_mg(x);
        acc_x = fTemp * 9.8f / 1000;
		fTemp = lsm303agr_from_fs_2g_hr_to_mg(y);
        acc_y = fTemp * 9.8f / 1000;
		fTemp = lsm303agr_from_fs_2g_hr_to_mg(z);
        acc_z = fTemp * 9.8f / 1000;
		
		printf("%s::,Accelerometer : X :%f m/s2, Y:%f m/s2, Z:%f m/s2\n",time_str,acc_x,acc_y,acc_z);
		
	}
	
	pthread_mutex_unlock(&bus_mutex);
}

#ifdef J1939_FUNCTION
//-------------------------
void get_j1939_data() {

	pthread_mutex_lock(&bus_mutex);
	int i;
	time_t mytime = time(NULL);
	char * time_str = ctime(&mytime);
	time_str[strlen(time_str)-1] = '\0';

	int ret = 0;
	struct J1939_message_t msg;
	uint32_t size = 0;
	memset(&msg, 0x00, sizeof(msg));
	size = 0;
	
	
	for(i = 0; i < AZ_VC_J1939_Read_Amount(); i++) //The amount of J1939 data received.
	{
		
		if((AZ_VC_J1939_Read(&msg, &size)) == 0) { 
			printf("%s::, J1939 Frame: ID:%X, Port:0, Priority:%x, DP:%x, PF:%X, PS:%X, SA:%x, PGN:%X, DLC:%x\n",time_str,msg.id, msg.p, msg.dp, msg.pf, msg.ps, msg.sa, msg.pgn, msg.dlc);
			printf("%s::, J1939 Data : %02x %02x %02x %02x %02x %02x %02x %02x\n",time_str, msg.data[0], msg.data[1], msg.data[2], msg.data[3], msg.data[4], msg.data[5], msg.data[6], msg.data[7]);
		}
		else
		{
			printf("ERROR:Failed to Read message from CAN Port.\n");
		}
	}
	
	pthread_mutex_unlock(&bus_mutex);
}
#else
//-------------------------
void get_j1939_data() {

	pthread_mutex_lock(&bus_mutex);
	int i;
	time_t mytime = time(NULL);
	char * time_str = ctime(&mytime);
	time_str[strlen(time_str)-1] = '\0';

	int ret = 0;
	struct can_message_t msg;
	uint32_t size = 0;
	memset(&msg, 0x00, sizeof(msg));
	size = 0;
	
	
	for(i = 0; i < AZ_VC_CAN_Read_Amount(); i++) //The amount of can data received.
	{
		
		if((AZ_VC_CAN_Read(&msg, &size)) == 0) { 
			printf("%s::, CAN Frame: ID:%X, Port:%x, Mode:%x, RTR:%x, DLC:%x\n",time_str,msg.id, msg.port, msg.mode, msg.rtr, msg.dlc);
			printf("%s::, CAN Data : %02x %02x %02x %02x %02x %02x %02x %02x\n",time_str, msg.data[0], msg.data[1], msg.data[2], msg.data[3], msg.data[4], msg.data[5], msg.data[6], msg.data[7]);
		}
		else
		{
			//printf("ERROR:Failed to Read message from CAN Port.\n");
		}
	}
	pthread_mutex_unlock(&bus_mutex);
}
#endif
//-------------------------
static void *read_can_sensor(void *args)
{
	while(stop_thread != 1)
	{
		get_sensor_data();
		get_j1939_data();
		usleep(100000); //100ms
	}
	pthread_exit(NULL);
}

//-------------------------
static void *request_sensor(void *args)
{

	while(stop_thread != 1) {
		request_acc_data();
		sleep(1);
	}

	pthread_exit(NULL);
}

//-------------------------
void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
		exit(0);
	}

}

//-------------------------
int main()
{
	//char port_name[] = {"/dev/faro_node"};
	char port_name[] = {"/dev/ttyUSB0"};
	int count = 0;
	struct can_fw_ver_t canfwver;
	struct sensor_ctrl_format_t sensor_ctrl;
	int ret =0;

	if(signal(SIGINT, sig_handler) == SIG_ERR)
	fprintf(stderr, "fail to catch SIGINT\n");


	// initialize mutex
	pthread_mutex_init(&bus_mutex, NULL);

	//if(FARO_CAN_SDK_DEBUG == 1)
	//fprintf(stdout, "starting get firmware version on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
	}


	// initialize
	if((ret = AZ_VC_Init(port_name)) == 0)
	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
		}

		fprintf(stdout, "get firmware version : v%02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}
	else
	{
		fprintf(stderr, "Calling AZ_VC_Init fail\n");	
		return 0;
	}	
#ifdef J1939_FUNCTION
    /*Activate the mode 2*/
	ret = AZ_VC_ModeActive(0,2);
#else
	/*Activate the mode 0*/
	ret = AZ_VC_ModeActive(0,0);
#endif
	if(ret != 0)
	{
		printf("ERROR: Failed to switch mode");
	}

	/*Configure the BaudRate*/
	struct can_config_t canCFG;
	canCFG.port = 0;
	canCFG.speed = 3;
	ret = AZ_VC_CAN_Config(canCFG);
	fprintf(stdout,"ret = %d\n",ret);

	if((ret = pthread_create(&do_read_can_sensor_thread, NULL, read_can_sensor, NULL)) != 0) {
		fprintf(stderr, "create do_read_can thread fail\n");
	}

	if((ret = pthread_create(&do_request_sensor_thread, NULL, request_sensor, NULL)) != 0) {
		fprintf(stderr, "create do_read_sensor thread fail\n");
	}

	pthread_join(do_read_can_sensor_thread, NULL);
	pthread_join(do_request_sensor_thread, NULL);
	
	return 0;
}
