/***********************************************************************
 *
 *  Copyright (c) 2017 - 2018 ANTZER TECH CO., LTD.
 *  All Rights Reserved
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/limits.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "faro_can_sdk_demo_j1708_config.h"
#include "faro_can_sdk.h"

#define FARO_CAN_SDK_J1708_WRITE_ARGS					9
#define FARO_CAN_SDK_J1708_READ_ARGS						1
#define FARO_CAN_SDK_J1708_FILTER_CONFIG_ARGS			3
#define FARO_CAN_SDK_J1708_GET_FILTER_COUNTS_ARGS		1
#define FARO_CAN_SDK_J1708_FILTER_GET_ARGS				1

#define FARO_CAN_SDK_CAN_CONFIG_ARGS					2

#define fprintf(fmt,...) {if (!g_quiet_print_flag) printf(__VA_ARGS__);}

enum J1708_TEST_IDX
{
	J1708_TEST_IDX_READ = 0,
	J1708_TEST_IDX_WRITE,
	J1708_TEST_IDX_FILTER_CONFIG,
	J1708_TEST_IDX_GET_FILTER_COUNTER,
	J1708_TEST_IDX_FILTER_GET,
};

struct J1708_filter_config_t
{
	uint8_t type;
	uint8_t mid;
};

static struct J1708_message_t j1708_send_msg;
static struct J1708_filter_config_t j1708_filter_config;

static int deinit_in_seconds = 10;
static int stop_thread = 0;
static char *port_name = NULL;
static uint8_t get_can_fw_ver = 0;
static uint8_t scan_port = 0;
static int test_cnt = 1;
static uint8_t mode_active = 0;
static int mode_active_port = 0;
static int mode_active_arg = 1;
static uint8_t launch_receiver = 0;
static uint8_t j1708_test = 0;
static int j1708_test_arg = 0;
static int current_active_mode = 0;
static uint8_t reset_module = 0;

static void display_help(int argc, char *argv[])
{
	fprintf(stdout, "Usage: %s [OPTION]\n"
			"\n"
			"  -h, --help\n"
			"  -s, --scan, scan available serial port\n"
			"  -p, --port\n"
			"		#1-port, ex \"/dev/ttyUSB0\"\n"
			"  -d, --deinit\n"
			"		#1-deinit after x seconds, -1 endless until ctrl-c\n"
			"  -g, --get_fw_ver, get CAN firmware version\n"
			"  -b, --reset_module\n"
			"  -m, --mode_active\n"
			"		#1-port: 0 or 1\n"
			"		#2-mode active: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR_OBD2, 5:CAN2ADR_J1939, 8:J1708_21, 9: J1708_21_OFF\n"
			"  -j, --j1708_test\n"
			"		0 - read\n"
			"		1 - write\n"
			"			#1-mid: 1 byte, ex: 0x00\n"
			"			#2-priority: 1 byte, ex: 0x01\n"
			"			#3-dlc: 1 byte, ex: 0x02\n"
			"			#4-4/19 data: 4 byte, ex: 0x01020304\n"
			"			#5-4/19 data: 4 byte, ex: 0x05060708\n"
			"			#6-4/19 data: 4 byte, ex: 0x090a0b0c\n"
			"			#7-4/19 data: 4 byte, ex: 0x0d0e0f10\n"
			"			#8-3/19 data: 3 byte, ex: 0x111213ff\n"
			"			    --> data will be combined as 0x0102030405060708090a0b0c0d0e0f10111213\n"
			"		2 - filter config\n"
			"			#1-type: 1 byte\n"
			"				0x01 : set filter\n"
			"				0xFF : reset all filter\n"
			"			#2-mid: 1 byte, ex: 0x00\n"
			"		3 - get filter count\n"
			"		4 - filter get\n"
			"  -t, --test_cnt, set counter for testing loop\n"
			"  -v, --version\n"
			"  -q, --quiet mode, set 1 to quiet mode\n"
			"\n",
			argv[0]);
	exit(0);
}

static int process_options(int argc, char *argv[])
{
	uint32_t tmp = 0;

	if(argc < 2) {
		display_help(argc, argv);
		return -1;
	}

	for (;;) {
		int option_index = 0;
		static const char *short_options = "q:hp:sd:gm:j:lt:vb";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"port", required_argument, 0, 'p'},
			{"scan", no_argument, 0, 's'},
			{"deinit", required_argument, 0, 'd'},
			{"get_fw_ver", no_argument, 0, 'g'},
			{"mode_active", required_argument, 0, 'm'},
			{"j1708_test", required_argument, 0,'j'},
			{"launch_receiver", no_argument, 0, 'l'},
			{"test_cnt", required_argument, 0, 't'},
			{"version", no_argument, 0, 'v'},
			{"quiet_mode", required_argument, 0, 'q'},
			{"reset_module", no_argument, 0, 'b'},
			{0,0,0,0},
		};

		int c = getopt_long(argc, argv, short_options, long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 'h':
			display_help(argc, argv);
			break;

		case 'p':
			port_name = strdup(optarg);
			break;

		case 's':
			scan_port = 1;
			break;

		case 'd':
			deinit_in_seconds = atoi(optarg);
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "deinit after %d seconds\n", deinit_in_seconds);
			break;

		case 'g':
			get_can_fw_ver = 1;
			break;

		case 'b':
			reset_module = 1;
			break;

		case 'm':
		{
			int cnt = 0;
			int index = optind - 1;
			int arg_num = 0;
			while(index < argc) {
				if(argv[index][0] == '-') {
					break;
				}

				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);
				
				switch(cnt) {
					case 0:
						if(atoi(argv[index]) < 0 || atoi(argv[index]) > 1)
						{
							fprintf(stdout,
							"invalid argument number.\nPort: 0 or 1\nMode: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR OBD2, 5:CAN2ADR J1939, 8:J1708_21 9:J1708_21_OFF\n");
							return -1;
						}
						else
						{
							mode_active_port = atoi(argv[index]);
						}
						break;

					case 1:
						if(atoi(argv[index]) < 0 || atoi(argv[index]) > 100)
						{
							fprintf(stdout,
							"invalid argument number.\nPort: 0 or 1\nMode: 0:RAW CAN, 1:OBD2, 2:J1939, 3:J1708_27byte, 4:CAN2ADR OBD2, 5:CAN2ADR J1939, 8:J1708_21 9:J1708_21_OFF\n");
							return -1;
						}
						else
						{
							mode_active_arg = atoi(argv[index]);
							mode_active = 1;
						}
						break;

					//default:
						
				}
				
				index++;
				cnt++;
				
				arg_num = FARO_CAN_SDK_CAN_CONFIG_ARGS;
				
				/*
				if(atoi(optarg) != 0 
					&& atoi(optarg) != 1 
					&& atoi(optarg) != 2 
					&& atoi(optarg) != 3 
					&& atoi(optarg) != 4
					&& atoi(optarg) != 5
					&& atoi(optarg) != 6
					&& atoi(optarg) != 8
					&& atoi(optarg) != 9) {
						fprintf(stderr, 
							"invalid argument number.\n0: RAW CAN, 1: OBD2, 2: J1939, 3: J1708_27byte, 4: CAN2ADR OBD2, 5: CAN2ADR J1939, 8: J1708_21 byte, 9: J1708_21 OFF\n");
						return -1;
				}
				else
				{
					mode_active_arg = atoi(optarg);
					mode_active = 1;
				}
				*/
			}
			if(cnt != arg_num)
				goto m_err;

			optind = index - 1;
			goto m_ok;

m_err:
			fprintf(stderr, "invalid argument number, should be %d\n", arg_num);

m_ok:
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "checking -m arguments done\n");
			break;
		}

		case 'j':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				j1708_test = 1;

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0:
							j1708_test_arg = atoi(argv[index]);
							if(j1708_test_arg < 0 ||
								j1708_test_arg > J1708_TEST_IDX_FILTER_GET)
								goto j_err;

							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								memset(&j1708_send_msg, 0x00, sizeof(j1708_send_msg));
							} else if(j1708_test_arg == J1708_TEST_IDX_FILTER_CONFIG) {
								memset(&j1708_filter_config, 0x00, sizeof(j1708_filter_config));
							}
							break;

						case 1:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1708_send_msg.mid = tmp;
							} else if(j1708_test_arg == J1708_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1708_filter_config.type = tmp;
							}
							break;

						case 2:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1708_send_msg.p = tmp;
							} else if(j1708_test_arg == J1708_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1708_filter_config.mid = tmp;
							}
							break;

						case 3:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1708_send_msg.dlc = tmp;
							}
							break;

						case 4:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1708_send_msg.data[0], &tmp, sizeof(uint32_t));
							}
							break;

						case 5:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1708_send_msg.data[4], &tmp, sizeof(uint32_t));
							}
							break;

						case 6:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1708_send_msg.data[8], &tmp, sizeof(uint32_t));
							}
							break;

						case 7:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1708_send_msg.data[12], &tmp, sizeof(uint32_t));
							}
							break;

						case 8:
							if(j1708_test_arg == J1708_TEST_IDX_WRITE) {
								sscanf(argv[index], "0x%08X", &tmp);
								tmp = ntohl(tmp);
								memcpy(&j1708_send_msg.data[16], &tmp, sizeof(uint32_t) - 1);
							}
							break;

						default:
							goto j_err;
					}

					index++;
					cnt++;
				};

				switch(j1708_test_arg) {
					case J1708_TEST_IDX_WRITE:				arg_num = FARO_CAN_SDK_J1708_WRITE_ARGS;				break;
					case J1708_TEST_IDX_READ:				arg_num = FARO_CAN_SDK_J1708_READ_ARGS;					break;
					case J1708_TEST_IDX_FILTER_CONFIG:		arg_num = FARO_CAN_SDK_J1708_FILTER_CONFIG_ARGS;		break;
					case J1708_TEST_IDX_GET_FILTER_COUNTER:	arg_num = FARO_CAN_SDK_J1708_GET_FILTER_COUNTS_ARGS;	break;
					case J1708_TEST_IDX_FILTER_GET:			arg_num = FARO_CAN_SDK_J1708_FILTER_GET_ARGS;			break;
				}

				if(cnt != arg_num)
					goto j_err;

				optind = index - 1;
				goto j_ok;

j_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

j_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -j arguments done\n");
			}
			break;

		case 'l':
			launch_receiver = 1;
			break;

		case 't':
			test_cnt = atoi(optarg);
			if(test_cnt <= 0)
				test_cnt = 1;
			break;

		case 'v':
			fprintf(stdout, "%s Version %d.%d.%d\n",
				argv[0],
				FARO_CAN_SDK_DEMO_J1708_VERSION_MAJOR,
				FARO_CAN_SDK_DEMO_J1708_VERSION_MINOR,
				FARO_CAN_SDK_DEMO_J1708_VERSION_PATCH);
			return -1;

		case 'q':
			g_quiet_print_flag = atoi(optarg);
			sleep(1);
			if(g_quiet_print_flag > 1 || g_quiet_print_flag < 0)
			{
				fprintf(stdout, "quiet mode set error, 0:Normal mode, 1:Quiet mode\n");
				return -1;
			}
			break;
		}
	}

	return 0;
}

static void *deinit_wait(void *args)
{
	int i;

	if(deinit_in_seconds == -1) {
		while(stop_thread == 0)
			sleep(1);
	} else {
		for(i=0; i<deinit_in_seconds; i++) {
			if(stop_thread != 0)
				break;

			sleep(1);
		}
	}

	if(current_active_mode == 3) {
		// change mode to 0
		if(AZ_VC_ModeActive(0x00,0) != 0) {
			fprintf(stderr, "calling AZ_VC_ModeActive fail\n");
		}
	}

	if(stop_thread == 0) {
		stop_thread = 1;
		sleep(1);
	}

	AZ_VC_DeInit();

	pthread_exit(NULL);
}

void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
	}

	if(launch_receiver) {
		AZ_VC_CAN_Dump_Pkt_Cnt();
	}
}

static int do_scan_port(void)
{
	int ret = 0;
	char buf[PATH_MAX];
	int serial_port_num = 0;

	memset(buf, sizeof(buf), 0x00);
	AZ_VC_GetCommPort(buf, &serial_port_num);
	fprintf(stdout, "Found %d serial ports\n", serial_port_num);
	if(serial_port_num > 0)
		fprintf(stdout, "buf -> \n%s", buf);

	return ret;
}

static void *do_get_can_fw_ver(void *args)
{
	int ret = 0;
	int i;
	struct can_fw_ver_t can_fw_ver;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting get firmware version on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail, test_cnt = %d\n", i);
			continue;
		}

		fprintf(stdout, "test_cnt = %d, can_fw_ver: %02X.%02X.%02X.%02X.%02X\n",
			i+1, can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type,can_fw_ver.year, can_fw_ver.week);
		
	}

err:
	pthread_exit(NULL);
}

static void *do_mode_active(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_mode_active on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		if((ret = AZ_VC_ModeActive(mode_active_port, mode_active_arg)) != 0) {
			fprintf(stderr, "calling AZ_VC_ModeActive fail\n");
			goto err;
		}

		current_active_mode = mode_active_arg;

		fprintf(stdout, "test_cnt = %d, mode_active_arg = %d\n", i, mode_active_arg);
	}
	stop_thread = 1;

err:
	pthread_exit(NULL);
}

static void *do_launch_receiver(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_launch_receiver on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	// change mode to J1708 for all J1708 testing
	ret = AZ_VC_ModeActive(0x00,3);
	if(ret != 0) {
		fprintf(stderr, "Calling AZ_VC_ModeActive fail\n");
		goto err;
	}
	current_active_mode = 3;

	AZ_VC_CAN_Start_Pkt_Consumer(start_consume_pkt_on);

err:
	pthread_exit(NULL);
}

static void *do_j1708_test(void *args)
{
	int ret = 0;
	int j1708_index;
	int j1708_amount;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_j1708_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	// change mode to J1708 for all J1708 testing
	ret = AZ_VC_ModeActive(0x00,3);
	if(ret != 0) {
		fprintf(stderr, "Calling AZ_VC_ModeActive fail\n");
		goto err;
	}
	current_active_mode = 3;

	switch(j1708_test_arg) {
		case J1708_TEST_IDX_WRITE:
			{
				int read_packet_cnt = 0;
				struct J1708_message_t msg;
				uint32_t read_size = 0;
				ret = AZ_VC_J1708_Write(j1708_send_msg);
				if(ret != 0) {
					fprintf(stderr, "Calling AZ_VC_J708_Write fail\n");
					goto err;
				}			
			}
			break;

		case J1708_TEST_IDX_READ:
			{
				//struct J1708_message_t msg;
				//uint32_t read_size = 0;
				int read_packet_cnt = 0;

				while(stop_thread == 0) {
					struct J1708_message_t msg;
					uint32_t read_size = 0;
					memset(&msg, 0x00, sizeof(msg));
					read_size = 0;
					//ret = AZ_VC_J1708_Read(&msg, &read_size);
					//if(ret == 0) {
					ret = 0;
					j1708_amount = AZ_VC_J1708_Read_Amount();
					for(j1708_index = 0; j1708_index < j1708_amount; j1708_index++)
					{
						if(ret = AZ_VC_J1708_Read(&msg, &read_size) == 0)
						{
							int i;

							fprintf(stdout, "J1708 read size = %d, packet_cnt = %d\n", read_size, (read_packet_cnt + 1));
							fprintf(stdout, "\tmid = 0x%02X\n", msg.mid);
							fprintf(stdout, "\tp = 0x%02X\n", msg.p);
							fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);

							for(i=0; i<sizeof(msg.data); i++) {
								fprintf(stdout, "\tdata[%d] = 0x%02X\n", i, msg.data[i]);
							}

							read_packet_cnt++;
						}
					}
					//fprintf(stdout, "ret = %d", ret);

					usleep(100);
				}
			}
			break;

		case J1708_TEST_IDX_FILTER_CONFIG:
			{
				ret = AZ_VC_J1708_FilterConfig(j1708_filter_config.type, j1708_filter_config.mid);
				fprintf(stdout, "J1708 filter config, ret = %d\n", ret);
			}
			break;

		case J1708_TEST_IDX_GET_FILTER_COUNTER:
			{
				uint16_t count = 0;

				AZ_VC_J1708_GetFilterCount(&count);
				fprintf(stdout, "J1708 filter count = %d\n", count);
			}
			break;

		case J1708_TEST_IDX_FILTER_GET:
			{
				uint16_t count = 0;
				uint32_t *filter = NULL;

				AZ_VC_J1708_GetFilterCount(&count);
				fprintf(stdout, "J1708 filter count = %d\n", count);

				if(count > 0) {
					int i;

					filter = malloc(sizeof(uint32_t) * count);
					memset(filter, 0x00, sizeof(uint32_t) * count);

					AZ_VC_J1708_FilterGet(filter);

					for(i=0; i<count; i++) {
						fprintf(stdout, "filter[%d] = 0x%08X\n", i, filter[i]);
					}

					free(filter);
				}
			}
			break;

		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}

static void *do_reset_module(void *args)
{
	int ret = 0;
	int i;

	deinit_in_seconds = 60;
	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
	
	//do reset module 3 times
	for(i = 0; i < 3; i++)
	{	
		if(stop_thread != 0)
			break;
		AZ_VC_Reset_Module();
		sleep(5);
	}
	
	struct can_fw_ver_t can_fw_ver;
	
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling reset module fail.");
			continue;
		}
		fprintf(stdout, "reset module finish.\n");
	}
	
	stop_thread = 1;
err:
	pthread_exit(NULL);
}

static int do_functions(void)
{
	int ret = 0;
	pthread_t deinit_wait_thread;
	pthread_t do_get_can_fw_ver_thread;
	pthread_t do_mode_active_thread;
	pthread_t do_launch_receiver_thread;
	pthread_t do_j1708_test_thread;
	pthread_t do_reset_module_thread;

	fprintf(stdout, "version %d.%d.%d\n",
		FARO_CAN_SDK_DEMO_J1708_VERSION_MAJOR,
		FARO_CAN_SDK_DEMO_J1708_VERSION_MINOR,
		FARO_CAN_SDK_DEMO_J1708_VERSION_PATCH);

	if(scan_port) {
		if((ret = do_scan_port()) != 0) {
			fprintf(stderr, "do_scan_port fail\n");
		}

		goto err;
	}

	// spawn deinit_wait_thread if necessary
	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "spawning deinit_wait thread...\n");

	if((ret = pthread_create(&deinit_wait_thread, NULL, deinit_wait, NULL)) != 0) {
		fprintf(stderr, "create deinit_wait thread fail\n");
			goto err;
	}

	if(get_can_fw_ver) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_can_fw_ver thread...\n");

		if((ret = pthread_create(&do_get_can_fw_ver_thread, NULL, do_get_can_fw_ver, NULL)) != 0) {
			fprintf(stderr, "create do_get_can_fw_ver thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_can_fw_ver_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_can_fw_ver_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(mode_active) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_mode_active thread...\n");

		if((ret = pthread_create(&do_mode_active_thread, NULL, do_mode_active, NULL)) != 0) {
			fprintf(stderr, "create do_mode_active thread fail\n");
			goto err;
		}

		ret = pthread_join(do_mode_active_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_mode_active_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(launch_receiver) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_launch_receiver thread...\n");

		if((ret = pthread_create(&do_launch_receiver_thread, NULL, do_launch_receiver, NULL)) != 0) {
			fprintf(stderr, "create do_launch_receiver thread fail\n");
			goto err;
		}

		ret = pthread_join(do_launch_receiver_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_launch_receiver_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(j1708_test) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_j1708_test thread...\n");

		if((ret = pthread_create(&do_j1708_test_thread, NULL, do_j1708_test, NULL)) != 0) {
			fprintf(stderr, "create do_j1708_test thread fail\n");
			goto err;
		}

		ret = pthread_join(do_j1708_test_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_j1708_test error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(reset_module) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_reset_module thread...\n");

		if((ret = pthread_create(&do_reset_module_thread, NULL, do_reset_module, NULL)) != 0) {
			fprintf(stderr, "create do_reset_module thread fail\n");
			goto err;
		}

		ret = pthread_join(do_reset_module_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_reset_module_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

normal_exit:
	ret = pthread_join(deinit_wait_thread, NULL);
	if(ret) {
		fprintf(stderr, "deinit_wait thread error: %s\n", strerror(ret));
		goto err;
	}

	fprintf(stderr, "threads are stopped\n");
	fflush(stderr);

err:
	scan_port = 0;
	get_can_fw_ver = 0;
	mode_active = 0;
	launch_receiver = 0;
	j1708_test = 0;

	return ret;
}

int main(int argc, char *argv[])
{
	int ret;

	if(signal(SIGINT, sig_handler) == SIG_ERR)
		fprintf(stderr, "fail to catch SIGINT\n");

	if(process_options(argc, argv) == 0)
		do_functions();

    if(port_name != NULL)
    	free(port_name);

	return 0;
}
