/***********************************************************************
 *
 *  Copyright (c) 2017 - 2018 ANTZER TECH CO., LTD.
 *  All Rights Reserved
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <memory.h>

#include "faro_can_sdk_test_config.h"
#include "faro_can_sdk.h"

static uint8_t can_data_receiver = 0;
static int stop_thread = 0;

#if defined(FARO_CAN_SDK_DEBUG) && FARO_CAN_SDK_DEBUG
extern void faro_can_sdk_hexdump_array(char *str_title, uint8_t *pbuf, ssize_t buf_len);
#endif

static void display_help(int argc, char *argv[])
{
	fprintf(stdout, "Usage: %s [OPTION]\n"
			"\n"
			"  -h, --help\n"
			"  -c, --can_data_receiver, ctrl-c to exit\n"
			"  -v, --version\n"
			"\n",
			argv[0]);
	exit(0);
}

static int process_options(int argc, char *argv[])
{
	if(argc < 2) {
		display_help(argc, argv);
		return -1;
	}

	for (;;) {
		int option_index = 0;
		static const char *short_options = "hcv";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"can_data_receiver", no_argument, 0, 'c'},
			{"version", no_argument, 0, 'v'},
			{0, 0, 0, 0},
		};

		int c = getopt_long(argc, argv, short_options, long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 'h':
			display_help(argc, argv);
			break;

		case 'c':
			can_data_receiver = 1;
			break;

		case 'v':
			fprintf(stdout, "%s Version %d.%d.%d\n",
				argv[0],
				FARO_CAN_SDK_TEST_VERSION_MAJOR,
				FARO_CAN_SDK_TEST_VERSION_MINOR,
				FARO_CAN_SDK_TEST_VERSION_PATCH);
			return -1;
		}
	}

	return 0;
}

static int do_can_data_receiver(void)
{
	int ret = 0;
	key_t msgq_key;
	int msgq_id;
	struct can_msg_payload_t msg;
	int cnt_pkt_received = 0;

	// make node
	if(mknod(FARO_CAN_SDK_MSGQ_NODE, S_IFREG | 0666, 0) != 0) {
		if(errno != EEXIST) {
			fprintf(stderr, "create message node fail - %s\n", FARO_CAN_SDK_MSGQ_NODE);
			goto err;
		}
	}

	// create key for message queue
	if((msgq_key = ftok(FARO_CAN_SDK_MSGQ_NODE, FARO_CAN_SDK_MSGQ_QUEUE_ID)) == -1) {
		fprintf(stderr, "create message queue key fail - %s\n", FARO_CAN_SDK_MSGQ_NODE);
		goto err;
	}

	if((msgq_id = msgget(msgq_key, IPC_CREAT | 0644)) == -1) {
		fprintf(stderr, "create message queue id fail - %s\n", FARO_CAN_SDK_MSGQ_NODE);
		goto err;
	}

	while(stop_thread == 0) {
		memset(&msg, 0x00, sizeof(msg));

		if(msgrcv(msgq_id, &msg, sizeof(msg.data), 0, 0) == -1) {
			fprintf(stderr, "calling msgrcv fail - %s\n", FARO_CAN_SDK_MSGQ_NODE);
			goto exit_while_loop;
		}

		fprintf(stdout, "Received packet %d - %s -->\n",
			(cnt_pkt_received + 1), FARO_CAN_SDK_MSGQ_NODE);
#if defined(FARO_CAN_SDK_DEBUG) && FARO_CAN_SDK_DEBUG
		faro_can_sdk_hexdump_array("msg.data", msg.data, sizeof(msg.data));
#endif

		cnt_pkt_received++;
	}


exit_while_loop:
	ret = 0;
	goto success;

err:
	ret = -1;

success:
#if 0 // un-remove megq_id, leave messages in the system
	for(i=0; i<NUM_OF_MSGQ; i++) {
		if(msgctl(msgq_id[i], IPC_RMID, NULL) == -1)
			fprintf(stderr, "remove message queue id fail\n");	
	}
#endif

	return ret;
}

static int do_functions(void)
{
	int ret = 0;

	fprintf(stdout, "version %d.%d.%d\n",
		FARO_CAN_SDK_TEST_VERSION_MAJOR,
		FARO_CAN_SDK_TEST_VERSION_MINOR,
		FARO_CAN_SDK_TEST_VERSION_PATCH);

	if(can_data_receiver) {
		if(do_can_data_receiver() != 0)
			goto err;
	}

	fprintf(stderr, "threads are stopped\n");
	fflush(stderr);

err:
	return ret;
}

static void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
	}
}

int main(int argc, char *argv[])
{
	int ret;

	if(signal(SIGINT, sig_handler) == SIG_ERR)
		fprintf(stderr, "fail to catch SIGINT\n");

	if(process_options(argc, argv) == 0)
		do_functions();

	return 0;
}
