/***********************************************************************
 *
 *  Copyright (c) 2019 - 2020 ANTZER TECH CO., LTD.
 *  All Rights Reserved
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/limits.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>

#include "faro_can_sdk_demo_config.h"
#include "faro_can_sdk.h"

#define fprintf(fmt,...) {if (!g_quiet_print_flag) printf(__VA_ARGS__);}

#define FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_32_BITS		6
#define FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_16_BITS		8
#define FARO_CAN_SDK_FILTER_CONFIG_RESET_REMOVE			4
#define FARO_CAN_SDK_J1939_RW_ARGS						14
#define FARO_CAN_SDK_J1939_PORT_CONFIG_ARGS				2
#define FARO_CAN_SDK_J1939_FILTER_CONFIG_ARGS			4
#define FARO_CAN_SDK_J1939_GET_FILTER_COUNTER_ARGS		2
#define FARO_CAN_SDK_J1939_FILTER_GET_ARGS				2
#define FARO_CAN_SDK_CAN_CONFIG_ARGS					2
#define FARO_CAN_SDK_GET_FILTER_CONFIG					2

enum J1939_TEST_IDX
{
	J1939_TEST_IDX_RW = 0,
	J1939_TEST_IDX_PORT_CONFIG = 1,
	J1939_TEST_IDX_FILTER_CONFIG = 2,
	J1939_TEST_IDX_GET_FILTER_COUNTER = 3,
	J1939_TEST_IDX_FILTER_GET = 4,
};

struct j1939_filter_config_t
{
	uint8_t type;
	uint8_t port;
	uint32_t pgn;
};

static struct can_filter_config_t can_filter_config;
static struct J1939_send_msg_t j1939_send_msg;
static struct j1939_filter_config_t j1939_filter_config;
static struct can_filter_config_t can_filter_config;

static int deinit_in_seconds = (10 * 3);	//set deinit time
static int stop_thread = 0;
static char *port_name = NULL;
static uint8_t get_can_fw_ver = 0;
static uint8_t scan_port = 0;
static uint8_t can_config = 0;
static struct can_config_t can_config_data;
static int test_cnt = 1;
static uint8_t mode_active = 0;
static int mode_active_arg = 1;
static uint8_t filter_config = 0;
static uint8_t filter_config_raw = 0;
static uint8_t get_filter_config = 0;
static uint8_t read_write_raw = 0;
static uint8_t launch_receiver = 0;
static uint8_t j1939_test = 0;
static int j1939_test_arg = 0;
static int j1939_test_port = 0;
static uint8_t reset_module = 0;

static void display_help(int argc, char *argv[])
{
	fprintf(stdout, "Usage: %s [OPTION]\n"
			"\n"
			"  -h, --help\n"
			"  -s, --scan, scan available serial port\n"
			"  -p, --port\n"
			"		#1-port, ex \"/dev/ttyUSB0\"\n"
			"  -d, --deinit\n"
			"		#1-deinit after x seconds, -1 endless until ctrl-c\n"
			"  -g, --get_fw_ver, get CAN firmware version\n"
			"  -c, --can_config, can config\n"
			"		#1-port: 1 byte\n"
			"		#2-speed: 1 byte, 0: 1M, 1: 800K, 2: 500K, 3: 250K, 4: 200K, 5: 125K\n"
			"  -b, --reset_module\n"
			"  -m, --mode_active\n"
			"		#1-mode active: 0: RAW CAN, 1: OBDII, 2: J1939, 3: J1708\n"
			"  -f, --filter_config\n"
			"  -w, --filter_config_raw\n"
			"		#1-type: 0: ID+Mask, 1: ID List, 2: Remove, 3: Reset All\n"
			"		#2-port: 0, 1\n"
			"		#3-bank: 0 - 13\n"
			"		#4-mode: 0: 16 bits, 1: 32 bits\n"
			"		#5-filterId: 16/32 bits value in hex, ex: 0x0011/0x00112233\n"
			"		#6-filterId/filterMask: 16bits/32 bits in hex, ex: 0x1234/0x00112233\n"
			"		#7-filterMask: 16 bits in hex, ex: 0x0011\n"
			"		#8-filterMask: 16 bits in hex, ex: 0x0011\n"
			"  -x, --get_filter_config\n"
			"       #1 port: 0, 1\n"
			"       #2 bank: 0 - 13\n"
			"  -r, --read_write_raw\n"
			"  -j, --j1939_test\n"
			"		0 - write/read test\n"
			"			#1-port: 1 byte\n"
			"			#2-priority: 1 byte\n"
			"			#3-sa: 1 byte\n"
			"			#4-pgn: 4 bytes, ex: 0x00112233\n"
			"			#5-dlc: 1 byte\n"
			"			#6-data-1: 1 byte\n"
			"			#7-data-2: 1 byte\n"
			"			#8-data-3: 1 byte\n"
			"			#9-data-4: 1 byte\n"
			"			#10-data-5: 1 byte\n"
			"			#11-data-6: 1 byte\n"
			"			#12-data-7: 1 byte\n"
			"			#13-data-8: 1 byte\n"
			"		1 - port config\n"
			"			#1-port: 0 or 1\n"
			"		2 - filter config\n"
			"			#1-type: 1 byte\n"
			"				0x00: Add one filter\n"
			"				0x01: Add one filter\n"
			"				0x02: Remove one filter\n"
			"				0x03: Reset all filters\n"
			"			#2-port: 1 byte\n"
			"			#3-pgn: 4 byte, ex: 0x00112233\n"
			"		3 - get filter count\n"
			"			#1-port: 0 or 1\n"
			"		4 - filter get\n"
			"			#1-port: 0 or 1\n"
			"  -l, --launch_receiver, using -d X to set waiting time in seconds\n"
			"  -t, --test_cnt, set counter for testing loop\n"
			"  -v, --version\n"
			"  -q, --quiet mode, set 1 to quiet mode\n"
			"\n",
			argv[0]);
	exit(0);
}

static int process_options(int argc, char *argv[])
{
	unsigned int tmp = 0;

	if(argc < 2) {
		display_help(argc, argv);
		return -1;
	}

	for (;;) {
		int option_index = 0;
		static const char *short_options = "q:hp:sd:gc:bm:fw:x:rj:lt:v";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"port", required_argument, 0, 'p'},
			{"scan", no_argument, 0, 's'},
			{"deinit", required_argument, 0, 'd'},
			{"get_fw_ver", no_argument, 0, 'g'},
			{"can_config", required_argument, 0, 'c'},
			{"reset_module", no_argument, 0, 'b'},
			{"mode_active", required_argument, 0, 'm'},
			{"filter_config", no_argument, 0, 'f'},
			{"filter_config_raw", required_argument, 0, 'w'},
			{"get_filter_config", required_argument, 0, 'x'},
			{"read_write_raw", no_argument, 0, 'r'},
			{"j1939_test", required_argument, 0,'j'},
			{"launch_receiver", no_argument, 0, 'l'},
			{"test_cnt", required_argument, 0, 't'},
			{"version", no_argument, 0, 'v'},
			{"quiet mode", required_argument, 0, 'q'},
			{0,0,0,0},
		};

		int c = getopt_long(argc, argv, short_options, long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 'h':
			display_help(argc, argv);
			break;

		case 'p':
			port_name = strdup(optarg);
			break;

		case 's':
			scan_port = 1;
			break;

		case 'd':
			deinit_in_seconds = atoi(optarg);
			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "deinit after %d seconds\n", deinit_in_seconds);
			break;

		case 'g':
			get_can_fw_ver = 1;
			break;

		case 'c':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				can_config = 1;
				memset(&can_config_data, 0x00, sizeof(can_config_data));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0:
							can_config_data.port = atoi(argv[index]);
							if(can_config_data.port < 0 || can_config_data.port > 1)
								goto c_err;

							break;

						case 1:
							can_config_data.speed = atoi(argv[index]);
							if(can_config_data.speed < 0 || can_config_data.speed > 5)
								goto c_err;

							break;

						default:
							goto c_err;
					}

					index++;
					cnt++;
				};

				arg_num = FARO_CAN_SDK_CAN_CONFIG_ARGS;

				if(cnt != arg_num)
					goto c_err;

				optind = index - 1;
				goto c_ok;

c_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

c_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -c arguments done\n");
			}
			break;

		case 'b':
			reset_module = 1;
			break;

		case 'm':
			mode_active = 1;
			mode_active_arg = atoi(optarg);
			if(mode_active_arg != 0 && mode_active_arg != 1 && mode_active_arg != 2 && mode_active_arg != 3 && mode_active_arg != 99) {
				fprintf(stderr, "invalid argument, %d, should be 0, 1, 2 or 3\n", mode_active_arg);
				display_help(argc, argv);
			}
			break;

		case 'f':
			filter_config = 1;
			break;

		case 'w':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				filter_config_raw = 1;
				memset(&can_filter_config, 0x00, sizeof(can_filter_config));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0: // 0 - ID+Mask, 1 - ID List, 2 - Remove, 3 - Reset All
							can_filter_config.type = atoi(argv[index]);
							if(can_filter_config.type < 0 || can_filter_config.type > 3)
								goto w_err;

							break;

						case 1:
							can_filter_config.port = atoi(argv[index]);
							if(can_filter_config.port < 0 && can_filter_config.port > 1)
								goto w_err;

							break;

						case 2:
							can_filter_config.bank = atoi(argv[index]);
							if(can_filter_config.bank < 0 || can_filter_config.bank > 13)
								goto w_err;

							break;

						case 3:
							can_filter_config.mode = atoi(argv[index]);
							if(can_filter_config.mode < 0 || can_filter_config.mode > 1)
								goto w_err;

							break;

						case 4:
							if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
								sscanf(argv[index], "0x%08X", &can_filter_config.filterId);
							} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterId |= (tmp << 16);
							}
							break;

						case 5:
							if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
								sscanf(argv[index], "0x%08X", &can_filter_config.filterMask);
							} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterId |= tmp;
							}
							break;

						case 6:
							if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterMask |= (tmp << 16);
							}
							break;

						case 7:
							if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
								(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
								sscanf(argv[index], "0x%04X", &tmp);
								can_filter_config.filterMask |= tmp;
							}
							break;

						default:
							goto w_err;
					}

					index++;
					cnt++;
				};

				if((can_filter_config.type == 2) || (can_filter_config.type == 3)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RESET_REMOVE;
				} else if((can_filter_config.type == 0 && can_filter_config.mode == 1) ||
					(can_filter_config.type == 1 && can_filter_config.mode == 1)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_32_BITS;
				} else if((can_filter_config.type == 0 && can_filter_config.mode == 0) ||
					(can_filter_config.type == 1 && can_filter_config.mode == 0)) {
					arg_num = FARO_CAN_SDK_FILTER_CONFIG_RAW_ARGS_16_BITS;
				}

				if(cnt != arg_num)
					goto w_err;

				optind = index - 1;
				goto w_ok;

w_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

w_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -w arguments done\n");
			}
			break;

		case 'x':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				get_filter_config = 1;
				memset(&can_filter_config, 0x00, sizeof(can_filter_config));

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					switch(cnt) {
						case 0: // port
							can_filter_config.port = atoi(argv[index]);
							if(can_filter_config.port < 0 || can_filter_config.port > 1)
								goto x_err;

							break;

						case 1: // bank
							can_filter_config.bank = atoi(argv[index]);
							if(can_filter_config.bank < 0 || can_filter_config.bank > 13)
								goto x_err;

							break;

						default:
							goto x_err;
					}

					index++;
					cnt++;
				};

				arg_num = FARO_CAN_SDK_GET_FILTER_CONFIG;
				if(cnt != arg_num)
					goto x_err;

				optind = index - 1;
				goto x_ok;

x_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

x_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -x arguments done\n");
			}
			break;

		case 'r':
			read_write_raw = 1;
			break;

		case 'j':
			{
				int cnt = 0;
				int index = optind - 1;
				int arg_num = 0;

				j1939_test = 1;

				while(index < argc) {
					if(argv[index][0] == '-') {
						break;
					}

					if(FARO_CAN_SDK_DEBUG == 1)
						fprintf(stdout, "argv[%d] = %s\n", index, argv[index]);

					tmp = 0;

					switch(cnt) {
						case 0: // 0: write/read, 1: port config, 2: filter config, 3: get port count
							j1939_test_arg = atoi(argv[index]);
							if(j1939_test_arg < 0 || j1939_test_arg > 4)
								goto j_err;

							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								memset(&j1939_send_msg, 0x00, sizeof(j1939_send_msg));
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								memset(&j1939_filter_config, 0x00, sizeof(j1939_filter_config));
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_GET) {
								j1939_test_port = 0;
							}

							break;

						case 1:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.port = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_PORT_CONFIG) {
								j1939_test_port = atoi(argv[index]);
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_filter_config.type = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_GET_FILTER_COUNTER) {
								j1939_test_port = atoi(argv[index]);
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_GET) {
								j1939_test_port = atoi(argv[index]);
							}
							break;

						case 2:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.p = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_filter_config.port = tmp;
							}
							break;

						case 3:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.sa = tmp;
							} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
								sscanf(argv[index], "0x%08X", &tmp);
								j1939_filter_config.pgn = tmp;
							}
							break;

						case 4:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%08X", &tmp);
								j1939_send_msg.pgn = tmp;
							}
							break;

						case 5:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.dlc = tmp;
							}
							break;

						case 6:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%08X", &tmp);
								j1939_send_msg.data[0] = tmp;
							}
							break;

						case 7:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[1] = tmp;
							}
							break;

						case 8:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[2] = tmp;
							}
							break;

						case 9:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[3] = tmp;
							}
							break;

						case 10:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[4] = tmp;
							}
							break;

						case 11:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[5] = tmp;
							}
							break;

						case 12:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[6] = tmp;
							}
							break;

						case 13:
							if(j1939_test_arg == J1939_TEST_IDX_RW) {
								sscanf(argv[index], "0x%02X", &tmp);
								j1939_send_msg.data[7] = tmp;
							}
							break;

						default:
							goto j_err;
					}

					index++;
					cnt++;
				};

				if(j1939_test_arg == J1939_TEST_IDX_RW) {
					arg_num = FARO_CAN_SDK_J1939_RW_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_PORT_CONFIG) {
					arg_num = FARO_CAN_SDK_J1939_PORT_CONFIG_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_CONFIG) {
					arg_num = FARO_CAN_SDK_J1939_FILTER_CONFIG_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_GET_FILTER_COUNTER) {
					arg_num = FARO_CAN_SDK_J1939_GET_FILTER_COUNTER_ARGS;
				} else if(j1939_test_arg == J1939_TEST_IDX_FILTER_GET) {
					arg_num = FARO_CAN_SDK_J1939_FILTER_GET_ARGS;
				}

				if(cnt != arg_num)
					goto j_err;

				optind = index - 1;
				goto j_ok;

j_err:
				fprintf(stderr, "invalid argument number, should be %d\n", arg_num);
				display_help(argc, argv);

j_ok:
				if(FARO_CAN_SDK_DEBUG == 1)
					fprintf(stdout, "checking -j arguments done\n");
			}
			break;

		case 'l':
			launch_receiver = 1;
			break;

		case 't':
			test_cnt = atoi(optarg);
			if(test_cnt <= 0)
				test_cnt = 1;
			break;

		case 'v':
			fprintf(stdout, "%s Version %d.%d.%d\n",
				argv[0],
				FARO_CAN_SDK_DEMO_VERSION_MAJOR,
				FARO_CAN_SDK_DEMO_VERSION_MINOR,
				FARO_CAN_SDK_DEMO_VERSION_PATCH);
			return -1;

		case 'q':
			g_quiet_print_flag = atoi(optarg);
			sleep(1);
			if(g_quiet_print_flag > 1 || g_quiet_print_flag < 0)
			{
				fprintf(stdout, "quiet mode set error, 0:Normal mode, 1:Quiet mode\n");
				return -1;
			}
			break;
		}
	}

	return 0;
}

static void *deinit_wait(void *args)
{
	int i;

	if(deinit_in_seconds == -1) {
		while(stop_thread == 0)
			sleep(1);
	} else {
		for(i=0; i<deinit_in_seconds; i++) {
			if(stop_thread != 0)
				break;

			sleep(1);
		}
	}

	if(stop_thread == 0) {
		stop_thread = 1;
		sleep(1);
	}

	AZ_VC_DeInit();

	pthread_exit(NULL);
}

void sig_handler(int signo)
{
	if(signo == SIGINT) {
		fprintf(stderr, "catch SIGINT\n");
		stop_thread = 1;
	}

	if(launch_receiver) {
		AZ_VC_CAN_Dump_Pkt_Cnt();
	}
}

static int do_scan_port(void)
{
	int ret = 0;
	char buf[PATH_MAX];
	int serial_port_num = 0;

	memset(buf, sizeof(buf), 0x00);
	AZ_VC_GetCommPort(buf, &serial_port_num);
	fprintf(stdout, "Found %d serial ports\n", serial_port_num);
	if(serial_port_num > 0)
		fprintf(stdout, "buf -> \n%s", buf);

	return ret;
}

static void *do_get_can_fw_ver(void *args)
{
	int ret = 0;
	int i;
	struct can_fw_ver_t can_fw_ver;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting get firmware version on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail, test_cnt = %d\n", i);
			continue;
		}

		fprintf(stdout, "get FW count = %d, can_fw_ver: %02X.%02X.%02X\n",
				i+1, can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}

err:
	pthread_exit(NULL);
}

static void *do_reset_module(void *args)
{
	int ret = 0;
	int i;

	deinit_in_seconds = 90;
	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;
	
	struct can_filter_config_t can_filter_config;
	memset(&can_filter_config, 0x00, sizeof(can_filter_config));
	
	can_filter_config.type = 0x03;
	can_filter_config.port = 0x00;
	can_filter_config.bank = 0x07;
	can_filter_config.mode = 0x00;
	AZ_VC_CAN_Filter_Config(can_filter_config);
	
	struct can_fw_ver_t can_fw_ver;

	memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
	
	//do reset module 3 times
	for(i = 0; i < 3; i++)
	{	
		if(stop_thread != 0)
			break;
		AZ_VC_Reset_Module();
		sleep(5);
	}
	
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling reset module fail.");
			continue;
		} 
		
		fprintf(stdout, "reset module finish.\n");
	}	
	stop_thread = 1;
err:
	pthread_exit(NULL);
}

static void *do_can_config(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting can_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		if((ret = AZ_VC_CAN_Config(can_config_data)) != 0)
			goto err;

		fprintf(stdout, "test_cnt = %d, port = %d, speed = %d\n",
			i, can_config_data.port, can_config_data.speed);
	}

err:
	stop_thread = 1;
	pthread_exit(NULL);
}

static void *do_mode_active(void *args)
{
	int ret = 0;
	int i;
	static int current_active_mode = 0;

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		if(port_name == NULL) {
			fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
			goto err;
		}
/*
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
*/
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;
		
		if((ret = AZ_VC_ModeActive(mode_active_arg)) != 0) {
			fprintf(stderr, "calling AZ_VC_ModeActive fail\n");
			goto err;
		}

		fprintf(stdout, "test_cnt = %d, mode_active_arg = %d\n", i, mode_active_arg);
	}

err:
	pthread_exit(NULL);
}

static void *do_filter_config(void *args)
{
	int ret = 0;
	int i, j;
	struct can_filter_config_t can_filter_config[] = {
		/*{2, 0, 0, 0, 0, 0},
		{2, 1, 0, 0, 0, 0},
		{2, 2, 0, 0, 0, 0},
		{2, 3, 0, 0, 0, 0},
		{2, 4, 0, 0, 0, 0},
		{2, 5, 0, 0, 0, 0},
		{2, 6, 0, 0, 0, 0},
		{2, 7, 0, 0, 0, 0},
		{2, 8, 0, 0, 0, 0},
		{2, 9, 0, 0, 0, 0},
		{2, 10, 0, 0, 0, 0},
		{2, 11, 0, 0, 0, 3},
		{2, 12, 0, 0, 0, 2},
		{2, 13, 0, 0, 0, 1},*/
		{2, 0, 0, 0, 0, 0},
		{2, 0, 1, 0, 0, 0},
		{2, 0, 2, 0, 0, 0},
		{2, 0, 3, 0, 0, 0},
		{2, 0, 4, 0, 0, 0},
		{2, 0, 5, 0, 0, 0},
		{2, 0, 6, 0, 0, 0},
		{2, 0, 7, 0, 0, 0},
		{2, 0, 8, 0, 0, 0},
		{2, 0, 9, 0, 0, 0},
		{2, 0, 10, 0, 0, 0},
		{2, 0, 11, 0, 0, 3},
		{2, 0, 12, 0, 0, 2},
		{2, 0, 13, 0, 0, 1},
	};

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_filter_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}

	for(i=0; i<test_cnt; i++) {
		for(j=0; j<(sizeof(can_filter_config)/sizeof(struct can_filter_config_t)); j++) {
			if(stop_thread != 0)
				break;

			if((ret = AZ_VC_CAN_Filter_Config(can_filter_config[j])) != 0)
				goto err;

			fprintf(stdout, "test_cnt = %d, type = %d, port = %d, bank = %d, filterId = %d, filterMask = %d\n",
				i, can_filter_config[j].type, can_filter_config[j].port, can_filter_config[j].bank, can_filter_config[j].filterId, can_filter_config[j].filterMask);
		}
	}

err:
	pthread_exit(NULL);
}

static void *do_filter_config_raw(void *args)
{
	int ret = 0;
	struct can_filter_config_t *filter = (struct can_filter_config_t *)args;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "type = %d, port = %d, bank = %d, mode = %d, filterId = 0x%08X, filerMask = 0x%08X\n",
			filter->type, filter->port, filter->bank, filter->mode, filter->filterId, filter->filterMask);

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_filter_config_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

/*
	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}
*/

	if((ret = AZ_VC_CAN_Filter_Config(*filter)) != 0)
		goto err;

err:
	pthread_exit(NULL);
}

static void *do_get_filter_config(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_get_filter_config on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}

	for(i=0; i<test_cnt; i++) {
		if(stop_thread != 0)
			break;

		fprintf(stdout, "set port = %d, bank = %d\n",
			can_filter_config.port, can_filter_config.bank);

		if((ret = AZ_VC_CAN_Filter_Get(&can_filter_config)) != 0)
			goto err;

		fprintf(stdout, "test_cnt = %d:\n", i);
		fprintf(stdout, "\ttype = %d, port = %d, bank = %d, mode = %d\n",
			can_filter_config.type, can_filter_config.port, can_filter_config.bank, can_filter_config.mode);
		fprintf(stdout, "\tfilterId = 0x%08X, filterMask = 0x%08X\n",
			can_filter_config.filterId, can_filter_config.filterMask);
	}

err:
	pthread_exit(NULL);
}

static void *do_read_write_raw(void *args)
{
	int ret = 0;
	int i, j, k;
	// workable package which is provided by Alex,
	// 10 02 81 01 00 00 00 01 00 00 08 11 22 33 44 55 66 77 88 10 03
	// dis-integrate
	// 10 02
	// 81				// cmd-id
	// 01				// port
	// 00 00 00 01		// id, 4 bytes
	// 00				// mode
	// 00				// rtr
	// 08				// dlc
	// 11 22 33 44 55 66 77 88	// data, 8 bytes
	// 10 03

	struct can_message_t send_msg[] = {
		{
			.port = 0x00,
			.id = 0x00000001,
			.mode = 0x01,
			.rtr = 0x00,
			.dlc = 0x08,
			.data[0] = 0x11,
			.data[1] = 0x22,
			.data[2] = 0x33,
			.data[3] = 0x44,
			.data[4] = 0x55,
			.data[5] = 0x66,
			.data[6] = 0x77,
			.data[7] = 0x88,
		},
		{
			.port = 0x01, 
			.id = 0x00000001,
			.mode = 0x00,
			.rtr = 0x00,
			.dlc = 0x08,
			.data[0] = 0x99,
			.data[1] = 0x88,
			.data[2] = 0x77,
			.data[3] = 0x66,
			.data[4] = 0x55,
			.data[5] = 0x44,
			.data[6] = 0x33,
			.data[7] = 0x22,
		},
	};

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_read_write_raw on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}


	if((ret = AZ_VC_ModeActive(0x00)) != 0)
		goto err;

	for(i=0; i<test_cnt; i++) {
		for(j=0; j<(sizeof(send_msg)/sizeof(struct can_message_t)); j++) {
			if(stop_thread != 0)
				break;

			if((ret = AZ_VC_CAN_Write(send_msg[j])) != 0)
				goto err;

			fprintf(stdout, "Send count = %d, port = %d, id = 0x%08X, data = ",	i+1, send_msg[j].port, send_msg[j].id);

			for(k=0; k<sizeof(send_msg[j].data); k++)
				fprintf(stdout, " 0x%02X", send_msg[j].data[k]);
			
			fprintf(stdout,"\n");

			send_msg[j].id++;


			// try to read something back
			struct can_message_t msg;
			uint32_t size;
			int k;

			memset(&msg, 0x00, sizeof(msg));
			size = 0;
				
			if((ret = AZ_VC_CAN_Read(&msg, &size)) == 0) {
				fprintf(stdout, "Receive count = %d, port = %d, ID = 0x%08X, data =", i+1, msg.port, msg.id);									
				for(k=0; k<sizeof(msg.data); k++)
					fprintf(stdout, " 0x%02X", msg.data[k]);
				fprintf(stdout,"\n");
				continue;
			}
			else
			{
				fprintf(stderr, "calling AZ_VC_CAN_Read fail\n");
				continue;
			}

			if(FARO_CAN_SDK_DEBUG == 1)
				fprintf(stdout, "test_cnt = %d, receive msg_id = 0x%08X, size=%d, msg->data --> \n", i+1, msg.id, size);

			if(FARO_CAN_SDK_DEBUG == 1) {
				for(k=0; k<sizeof(msg.data); k++)
					fprintf(stdout, "msg.data[%d] = 0x%02X\n", k, msg.data[k]);
			}
		}
	}

	stop_thread = 1;
err:
	pthread_exit(NULL);
}

static void *do_launch_receiver(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_launch_receiver on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	/*{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}*/

	AZ_VC_CAN_Start_Pkt_Consumer(start_consume_pkt_on);

err:
	pthread_exit(NULL);
}

static void *do_j1939_test(void *args)
{
	int ret = 0;
	int i;

	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "starting do_j1939_test on %s\n", port_name);

	if(port_name == NULL) {
		fprintf(stderr, "invalid port_name, assigned by -p /dev/ttyUSBx\n");
		goto err;
	}

	// initialize
	if((ret = AZ_VC_Init(port_name, BR)) != 0)
		goto err;

	{
		struct can_fw_ver_t can_fw_ver;

		memset(&can_fw_ver, 0x00, sizeof(can_fw_ver));
		if((ret = AZ_VC_GetFWVer(&can_fw_ver)) != 0) {
			fprintf(stderr, "Calling AZ_VC_GetFWVer fail\n");
			goto err;
		}

		fprintf(stdout, "can_fw_ver: %02X.%02X.%02X\n", can_fw_ver.main, can_fw_ver.minor, can_fw_ver.type);
	}

	// change mode to J1939 for all J1939 testing
	ret = AZ_VC_ModeActive(0x02);
	if(ret != 0) {
		fprintf(stderr, "Calling AZ_VC_ModeActive fail\n");
		goto err;
	}

	switch(j1939_test_arg) {
		case J1939_TEST_IDX_RW:
			{
				struct J1939_message_t msg;
				uint32_t read_size = 0;

				ret = AZ_VC_J1939_Write(j1939_send_msg);
				if(ret != 0) {
					fprintf(stderr, "Calling AZ_VC_J1939_Write fail\n");
					goto err;
				}				

				memset(&msg, 0x00, sizeof(msg));
				ret = AZ_VC_J1939_Read(&msg, &read_size);
				fprintf(stdout, "J1939 read size = %d\n", read_size);
				fprintf(stdout, "\tid = 0x%08X\n", msg.id);
				fprintf(stdout, "\tp = 0x%02X\n", msg.p);
				fprintf(stdout, "\tdp = 0x%02X\n", msg.dp);
				fprintf(stdout, "\tpf = 0x%02X\n", msg.pf);
				fprintf(stdout, "\tps = 0x%02X\n", msg.ps);
				fprintf(stdout, "\tsa = 0x%02X\n", msg.sa);
				fprintf(stdout, "\tpgn= 0x%02X\n", msg.pgn);
				fprintf(stdout, "\tdlc = 0x%02X\n", msg.dlc);

				for(i=0; i<8; i++) {
					fprintf(stdout, "\tdata[%d] = 0x%02X\n", i, msg.data[i]);
				}
			}
			break;

		case J1939_TEST_IDX_PORT_CONFIG:
			{
				ret = AZ_VC_J1939Port_Config(j1939_test_port);
				fprintf(stdout, "J1939 port config, ret = %d\n", ret);
			}
			break;

		case J1939_TEST_IDX_FILTER_CONFIG:
			{
				ret = AZ_VC_J1939Filter_Config(j1939_filter_config.type, j1939_filter_config.port, j1939_filter_config.pgn);
				fprintf(stdout, "J1939 filter config, ret = %d\n", ret);
			}
			break;

		case J1939_TEST_IDX_GET_FILTER_COUNTER:
			{
				uint16_t count = 0;

				AZ_VC_J1939_get_filter_count(j1939_test_port, &count);
				fprintf(stdout, "J1939 filter count = %d\n", count);
			}
			break;

		case J1939_TEST_IDX_FILTER_GET:
			{
				uint16_t count = 0;
				uint32_t *filter = NULL;

				AZ_VC_J1939_get_filter_count(j1939_test_port, &count);
				fprintf(stdout, "J1939 filter count = %d\n", count);

				if(count > 0) {
					int i;

					filter = malloc(sizeof(uint32_t) * count);
					memset(filter, 0x00, sizeof(uint32_t) * count);

					AZ_VC_J1939Filter_Get(j1939_test_port, filter);

					for(i=0; i<count; i++) {
						fprintf(stdout, "filter[%d] = 0x%08X\n", i, filter[i]);
					}

					free(filter);
				}
			}
			break;

		default:
			fprintf(stderr, "Invalid arguments\n");
			goto err;
	}

err:
	pthread_exit(NULL);
}

static int do_functions(void)
{
	int ret = 0;
	pthread_t deinit_wait_thread;
	pthread_t do_get_can_fw_ver_thread;
	pthread_t do_can_config_thread;
	pthread_t do_mode_active_thread;
	pthread_t do_filter_config_thread;
	pthread_t do_filter_config_raw_thread;
	pthread_t do_get_filter_config_thread;
	pthread_t do_read_write_raw_thread;
	pthread_t do_launch_receiver_thread;
	pthread_t do_j1939_test_thread;
	pthread_t do_reset_module_thread;

	fprintf(stdout, "version %d.%d.%d\n",
		FARO_CAN_SDK_DEMO_VERSION_MAJOR,
		FARO_CAN_SDK_DEMO_VERSION_MINOR,
		FARO_CAN_SDK_DEMO_VERSION_PATCH);

	if(scan_port) {
		if((ret = do_scan_port()) != 0) {
			fprintf(stderr, "do_scan_port fail\n");
		}

		goto err;
	}

	// spawn deinit_wait_thread if necessary
	if(FARO_CAN_SDK_DEBUG == 1)
		fprintf(stdout, "spawning deinit_wait thread...\n");

	if((ret = pthread_create(&deinit_wait_thread, NULL, deinit_wait, NULL)) != 0) {
		fprintf(stderr, "create deinit_wait thread fail\n");
			goto err;
	}

	if(get_can_fw_ver) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_can_fw_ver thread...\n");

		if((ret = pthread_create(&do_get_can_fw_ver_thread, NULL, do_get_can_fw_ver, NULL)) != 0) {
			fprintf(stderr, "create do_get_can_fw_ver thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_can_fw_ver_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_can_fw_ver_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(reset_module) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_reset_module thread...\n");

		if((ret = pthread_create(&do_reset_module_thread, NULL, do_reset_module, NULL)) != 0) {
			fprintf(stderr, "create do_reset_module thread fail\n");
			goto err;
		}

		ret = pthread_join(do_reset_module_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_reset_module_thread thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(can_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_can_config thread...\n");
		if((ret = pthread_create(&do_can_config_thread, NULL, do_can_config, NULL)) != 0) {
			fprintf(stderr, "create do_can_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_can_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_can_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(mode_active) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_mode_active thread...\n");
		if((ret = pthread_create(&do_mode_active_thread, NULL, do_mode_active, NULL)) != 0) {
			fprintf(stderr, "create do_mode_active thread fail\n");
			goto err;
		}

		ret = pthread_join(do_mode_active_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_mode_active_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(filter_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_filter_config thread...\n");
		if((ret = pthread_create(&do_filter_config_thread, NULL, do_filter_config, NULL)) != 0) {
			fprintf(stderr, "create do_filter_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_filter_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_filter_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	} 

	if(filter_config_raw) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_filter_config_raw thread...\n");
		if((ret = pthread_create(&do_filter_config_raw_thread, NULL, do_filter_config_raw, &can_filter_config)) != 0) {
			fprintf(stderr, "create do_filter_config_raw thread fail\n");
			goto err;
		}

		ret = pthread_join(do_filter_config_raw_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_filter_config_raw_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(get_filter_config) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_get_filter_config thread...\n");
		if((ret = pthread_create(&do_get_filter_config_thread, NULL, do_get_filter_config, NULL)) != 0) {
			fprintf(stderr, "create do_get_filter_config thread fail\n");
			goto err;
		}

		ret = pthread_join(do_get_filter_config_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_get_filter_config_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(read_write_raw) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_read_write_raw thread...\n");
		if((ret = pthread_create(&do_read_write_raw_thread, NULL, do_read_write_raw, NULL)) != 0) {
			fprintf(stderr, "create do_read_write_raw thread fail\n");
			goto err;
		}

		ret = pthread_join(do_read_write_raw_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_read_write_raw_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(launch_receiver) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawning do_launch_receiver thread...\n");
		if((ret = pthread_create(&do_launch_receiver_thread, NULL, do_launch_receiver, NULL)) != 0) {
			fprintf(stderr, "create do_launch_receiver thread fail\n");
			goto err;
		}

		ret = pthread_join(do_launch_receiver_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_launch_receiver_thread error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

	if(j1939_test) {
		if(FARO_CAN_SDK_DEBUG == 1)
			fprintf(stdout, "spawn do_j1939_test thread...\n");
		if((ret = pthread_create(&do_j1939_test_thread, NULL, do_j1939_test, NULL)) != 0) {
			fprintf(stderr, "create do_j1939_test thread fail\n");
			goto err;
		}

		ret = pthread_join(do_j1939_test_thread, NULL);
		if(ret) {
			fprintf(stderr, "do_j1939_test error: %s\n", strerror(ret));
			goto err;
		}

		goto normal_exit;
	}

normal_exit:
	ret = pthread_join(deinit_wait_thread, NULL);
	if(ret) {
		fprintf(stderr, "deinit_wait thread error: %s\n", strerror(ret));
		goto err;
	}


	fprintf(stderr, "threads are stopped\n");
	fflush(stderr);

err:
	scan_port = 0;
	get_can_fw_ver = 0;
	can_config = 0;
	mode_active = 0;
	filter_config = 0;
	filter_config_raw = 0;
	get_filter_config = 0;
	read_write_raw = 0;
	launch_receiver = 0;
	j1939_test = 0;

	return ret;
}

int main(int argc, char *argv[])
{
	int ret;

	if(signal(SIGINT, sig_handler) == SIG_ERR)
		fprintf(stderr, "fail to catch SIGINT\n");

	if(process_options(argc, argv) == 0)
		do_functions();

    if(port_name != NULL)
    	free(port_name);

	return 0;
}
